#!/usr/bin/python
# -*- coding: utf-8 -*-

import requests
import os
import platform
import sys
import time
import operator
import glob
import subprocess
import six
import psutil
import shutil
from argparse import ArgumentParser,_SubParsersAction
from sys import platform as _platform
from colorama import init as colorama_init, Fore, Style
from .asar import Asar
from string import Template


SLACK_PLUGIN_CODE = """
const { app } = require('electron')
const { BrowserWindow } = require('electron')

payload = `
function changeStyle() { 
    var classes = ['ql-editor', 'c-message__body', 'message_body', 'c-message_attachment__text', 'msg_inline_attachment_row', 'c-mrkdwn__pre', 'c-message_kit__text'];

    classes.forEach((cls) => {
      for (let item of document.getElementsByClassName(cls))
      { 
        item.setAttribute('dir','auto');
        item.style.textAlign = 'start';
      }
    });

    classes = ['c-message__edited_label'];

    classes.forEach((cls) => {
      for (let item of document.getElementsByClassName(cls))
      { 
        item.style.display = 'inline-block';
      }
    });
    

    //$$('.ql-editor, .c-message__body, .message_body, .c-message_attachment__text, .msg_inline_attachment_row, .c-mrkdwn__pre, .c-message_kit__text').attr('dir', 'auto').css('text-align', 'start');
    //$$('.c-message__edited_label').css('display','inline-block');
}
function doIt() {
  document.getElementsByTagName('body')[0].addEventListener('DOMSubtreeModified', changeStyle)
}

doIt();
`
$debug_script

app.on('web-contents-created', (evt, webContents) => {
     webContents.on('did-finish-load', function() {
        webContents.executeJavaScript(payload); 
     });
});
"""

SLACK_DEBUGGER_CODE = """
app.commandLine.appendSwitch('remote-debugging-port', '9222');
"""


def find_app_local_path(app):
    app_root = os.path.join(os.environ['LOCALAPPDATA'],app)
    if os.path.isdir(app_root): # Local install mode
        apps = os.path.join(app_root, "app-")
        candidates = [x for x in glob.glob(os.path.join(apps + "*")) if os.path.isdir(x)]

        versions = [[int(y) for y in x.rsplit('-',1)[-1].split('.')] for x in candidates]
        max_index, max_value = max(enumerate(versions), key=operator.itemgetter(1))
        return candidates[max_index]

def find_app_store_path(app):
    try:
        app_root = subprocess.check_output(["powershell.exe", "-NoLogo", "-NoProfile", "-NonInteractive", '(Get-AppxPackage | where {$_.Name -match \"'+app+'\"} ).InstallLocation'])
        app_root = app_root.decode('utf-8').strip()
        if not os.path.isdir(app_root): return None
        return os.path.join(app_root, app)
    except subprocess.CalledProcessError:
        pass


def find_app_path(location, app):
    if _platform == 'darwin':
        if location == "auto":
            p = '/Applications/' + app + '.app/Contents/MacOS'
        else: 
            p = location
        if not os.path.isdir(p) or not os.path.isfile(os.path.join(p, app)):
            raise Exception("%s is not a valid slack directory" % p)
        asar_path = os.path.join(p,"resources") 
        if len([name for name in os.listdir(p)]) is 1:
                asar_path = '/Applications/' + app + '.app/Contents/Resources' 
        return os.path.join(p, app), asar_path

    elif _platform == 'win32' or _platform == 'win64':
        if location == "store":
            p = find_app_store_path(app)
        elif location == "local":
            p = find_app_local_path(app)
        elif location == "auto":
            p = find_app_local_path(app)
            if p is None: p = find_app_store_path(app)
        else:
            p = location

        if p is None:
            raise Exception("Cannot find a valid {} path in {}".format(app, location))
        if not os.path.isdir(p) or not os.path.isfile(os.path.join(p, app + ".exe")):
            raise Exception("{} find a valid {} directory".format(p, app))
        return os.path.join(p, app + ".exe"), os.path.join(p,"resources")

    elif _platform.startswith('linux'):
        p = None
        if location == "auto":
            for path in os.environ["PATH"].split(os.pathsep):
                p = os.path.join(path, app)
                if os.path.isfile(p) and os.access(p, os.X_OK):
                    p=os.path.realpath(p)
                    path=os.path.dirname(p)
                    break
        else:
            p = location
            path = os.path.split(location)

        if p is None:
            raise Exception("Could not find %s in path" % app)

        if os.path.isfile(p) and os.access(p, os.X_OK):
            return p, os.path.join(path, "resources")
        else:
            raise Exception("{} find a valid {} path".format(p, app))
    else:
        raise Exception ("%s is not a supported platform" % _platfom)


def run_app(path, param):
    DETACHED_PROCESS = 0x00000008

    if _platform == 'darwin':
        cmd = path
        if param is not None:
            cmd = cmd + " " + param
        else:
            cmd = "open -a " + cmd
        subprocess.Popen([cmd], shell=True)

    elif _platform == 'win32' or _platform == 'win64':
        cmd = [path]
        if param is not None:
            cmd = cmd + [param]
        subprocess.Popen(cmd, creationflags=DETACHED_PROCESS, shell=True)

    elif _platform.startswith('linux'):
        if (param is not None):
            arg = [path, param]
        else:
            arg = path
        subprocess.Popen(arg, shell=False)

    else:
        raise Exception("%s is not a supported platform" % _platfom)

def kill_existing_app(app):
    for i in psutil.process_iter():               
        try:
            name = os.path.splitext(i.name())[0]
            if name.lower() == app:
                six.print_("Killing %s process Pid:%s%s%s." % (app, Fore.GREEN, i.pid, Style.RESET_ALL), end='\n', flush=True)
                try:
                    i.terminate()
                except:
                    pass
        except psutil.AccessDenied:
            pass

def edit_file(content, script, prefix):
    COMMENT_PRE = "\n// ### INSERTED BELOW ### //\n".encode('utf-8')
    COMMENT_POST = "\n// ### INSERTED ABOVE ### //\n".encode('utf-8')
    loc_pre = content.rfind(COMMENT_PRE)
    loc_post = content.rfind(COMMENT_POST)
    if (loc_pre != -1): 
        if (loc_post == -1): 
            raise Exception("found prefix at %d but can't find postfix" % loc_pre)
        content = content[:loc_pre] + content[loc_post:]
    if prefix:
        return COMMENT_PRE + script + COMMENT_POST + content
    else:
        return content + COMMENT_PRE + script + COMMENT_POST

def find_target_file_in_asar(asar):
    knowns = [
            os.path.join("dist","main.js")
    ]

    for f in knowns: 
        if f in asar: return f
    raise Exception("Can't find any of %s in asar file" % str(knowns))
    

def do_edit_method(args, app_path, asar_path):
    backup_file = os.path.join(asar_path, args.backup)
    backup_unpacked_path = os.path.join(asar_path, args.backup + ".unpacked")
    asar_file = os.path.join(asar_path, 'app.asar')
    asar_unpacked_path = os.path.join(asar_path, 'app.asar.unpacked')
    if os.path.exists(backup_file) or (os.path.exists(asar_unpacked_path) and os.path.exists(backup_unpacked_path)):
        if not args.force:
            raise Exception("Backup file already exists, consider using --force. Stopped" + backup_file)
    else:
        six.print_("Backup %s as %s%s%s." % ('app.asar', Fore.GREEN, backup_file, Style.RESET_ALL))
        shutil.copy(asar_file, backup_file)
        if os.path.exists(asar_unpacked_path):
            shutil.copytree(asar_unpacked_path, backup_unpacked_path)

    asar = Asar.open(asar_file)
    stat_file = find_target_file_in_asar(asar)

    template = Template(SLACK_PLUGIN_CODE)
    if args.debug:
        script = template.substitute(debug_script=SLACK_DEBUGGER_CODE)
    else:
        script = template.substitute(debug_script="")


    asar[stat_file] = edit_file(asar[stat_file], script.encode('utf-8'), True)
    six.print_("Marking %s%s%s in %s%s%s as unpacked file." % (Fore.GREEN, stat_file, Style.RESET_ALL, Fore.GREEN, asar_file, Style.RESET_ALL))
    asar.mark_packed(stat_file, True)
    asar.save(asar_file)
    six.print_("Patching %s%s%s." % (Fore.GREEN, asar_file, Style.RESET_ALL))

    if args.start:
        six.print_("Running %s from %s %s." % ('slack', Fore.GREEN, app_path))
        run_app(app_path, None)

def main():
    parser = ArgumentParser(description=""" 
    Inject hebrew support plugin into Slack's tab inside an electron app.

    This program injects the Chrome's hebrew_slack plugin into any electron (desktop) version of the slack app
    """)
    parser.add_argument("-l", "--location",
                      action="store", dest="location", default="auto",
                      help="Location of application to run, or auto, local (Windows only), store (Windows only) [default: auto]")

    parser.add_argument("--no-kill", dest="kill", action='store_false',
                        help="Do not attempt to kill original application before starting",
                        default=True
                        )

    parser.add_argument("--no-start", dest="start", action='store_false',
                        help="Do not attempt to start application (assume already running)",
                        default=True
                        )

    parser.add_argument("-b", "--backup", 
                         default = "app.asar.orig",
                         help = "Name to use save original slack app backup. This will never overwrite an existing backup file. Fails if file already exists and not used with -f [default: %(default)s]")
    parser.add_argument("-f", "--force", 
                         default = False,
                         action = "store_true",
                         help = "Proceed even if backup file already exists [default: %(default)s]")

    parser.add_argument("-d", "--debug",
                        default = False,
                        action = "store_true",
                        help = "Pass --remote-debugging-port=9222 to enable rendered debugger with chrome")

    # parse args
    args  = parser.parse_args()

    app_path, asar_path = find_app_path(args.location, 'slack')

    colorama_init(autoreset=True)

    if args.kill:
        kill_existing_app('slack')


    do_edit_method(args, app_path, asar_path)

    six.print_("Done")



import io
import os
import json
import errno
import struct
import shutil


def round_up(i, m):
    """Rounds up ``i`` to the next multiple of ``m``.

    ``m`` is assumed to be a power of two.
    """
    return (i + m - 1) & ~(m - 1)

def forward_path_split(path):
    return ([''] + os.path.normpath(path).split(os.path.sep,1) )[-2:]


class Asar:

    """Represents an asar file.

    You probably want to use the :meth:`.open` or :meth:`.from_path`
    class methods instead of creating an instance of this class.

    Attributes
    ----------
    path : str
        Path of this asar file on disk.
        If :meth:`.from_path` is used, this is just
        the path static_method to it.
    fp : File-like object
        Contains the data for this asar file.
    header : dict
        Dictionary used for random file access.
    base_offset : int
        Indicates where the asar file header ends.
    """

    def __init__(self, header):
        self.header = header

    @classmethod
    def open(cls, path):
        """Decodes the asar file from the given ``path``.

        You should use the context manager interface here,
        to automatically close the file object when you're done with it, i.e.

        .. code-block:: python

            with Asar.open('./something.asar') as a:
                a.extract('./something_dir')

        Parameters
        ----------
        path : str
            Path of the file to be decoded.
        """
        fp = open(path, 'rb')

        # decode header
        # NOTE: we only really care about the last value here.
        data_size, header_size, header_object_size, header_string_size = struct.unpack('<4I', fp.read(16))


        header_json = fp.read(header_string_size).decode('utf-8')
        header=json.loads(header_json)

        base_offset=round_up(16 + header_string_size, 4)

        def _inline_header(header, fp, base_offset, unpacked_path):
            new_dict = {"files": {} }
            for (k, v) in header['files'].items():
                if "files" in v:
                    new_dict['files'][k] = _inline_header(v, fp, base_offset, os.path.join(unpacked_path, k))
                elif "link" in v: #Ignore pre-set files
                    new_dict['files'][k] = {}
                    new_dict['files'][k].update(v)
                elif "unpacked" in v:
                    new_dict['files'][k] = {}
                    new_dict['files'][k].update(v)
                    new_dict['files'][k]['content'] = open(os.path.join(unpacked_path, k),'rb').read()
                    new_dict['files'][k].pop('size',None)
                else:
                    new_dict['files'][k] = {}
                    new_dict['files'][k].update(v)
                    fp.seek(base_offset + int(v['offset']))
                    new_dict['files'][k]['content'] = fp.read(v['size'])
                    new_dict['files'][k].pop('size',None)
                    new_dict['files'][k].pop('offset',None)
            return new_dict

        header = _inline_header(header, fp, base_offset, path + ".unpacked")
        return cls(header = header)

    @staticmethod
    def _build(header, concatenated_files):
        header_json = json.dumps(header, sort_keys=True, separators=(',', ':')).encode('utf-8')

        # TODO: using known constants here for now (laziness)...
        #       we likely need to calc these, but as far as discord goes we haven't needed it.
        header_string_size = len(header_json)
        data_size = 4 # uint32 size
        aligned_size = round_up(header_string_size, data_size)
        header_size = aligned_size + 8
        header_object_size = aligned_size + data_size

        # pad remaining space with NULLs
        diff = aligned_size - header_string_size
        header_json = header_json + b'\0' * (diff) if diff else header_json

        fp = io.BytesIO()
        fp.write(struct.pack('<4I', data_size, header_size, header_object_size, header_string_size))
        fp.write(header_json)
        fp.write(concatenated_files)

        return {'path':None,
                'fp':fp,
                'header':header,
                'base_offset':round_up(16 + header_string_size, 4),
                'unpacked_files' : {}
        }

    @classmethod
    def from_path(cls, path):
        """Creates an asar file using the given ``path``.

        When this is used, the ``fp`` attribute of the returned instance
        will be a :class:`io.BytesIO` object, so it's not written to a file.
        You have to do something like:

        .. code-block:: python

            with Asar.from_path('./something_dir') as a:
                with open('./something.asar', 'wb') as f:
                    a.fp.seek(0) # just making sure we're at the start of the file
                    f.write(a.fp.read())

        You cannot exclude files/folders from being packed yet.

        Parameters
        ----------
        path : str
            Path to walk into, recursively, and pack
            into an asar file.
        """

        def _path_to_dict(path):
            result = {'files': {}}

            for f in os.scandir(path):
                if os.path.isdir(f.path):
                    result['files'][f.name] = _path_to_dict(f.path)
                elif f.is_symlink():
                    result['files'][f.name] = {
                        'link': os.path.realpath(f.name)
                    }
                else:
                    with open(f.path, 'rb') as fp:
                        result['files'][f.name] = {
                            'content' : fp.read()
                        }


            return result

        header = _path_to_dict(path)
        return cls(header)

    def __setitem__(self, name, content):
        FILE_IS_UNPACKED = object()

        def _set_inlined_header(header, item, content):
            (p,n) = forward_path_split(item)
            if not 'files' in header:
                raise Exception("error modifying dictionary")

            if (p != ''):
                # 'files' in root key
                # Either P doesn't exist, or it exists and isn't a directory. 
                if not p in header['files'] or not 'files' in header['files'][p]:
                    header['files'][p] = {'files':{}}
                _set_inlined_header(header['files'][p], n, content)
            else: # we're at the target item
                header['files'][n]['content'] = content



        if (type(content) != bytes):
            raise Exception("content must be gives as a bytes object")
        _set_inlined_header(self.header, name, content)

    def __getitem__(self, item):
        def _rec_get_item(header, item):
            (p,n) = forward_path_split(item)
            if not 'files' in header:
                raise Exception("error in header")
            if p != '':
                if not p in header['files'] or not 'files' in header['files'][p]:
                    raise Exception("item not found")
                return _rec_get_item(header['files'][p], n)
            else: # we're at target item
                if 'content' in header['files'][n]: # First check if we already set this file
                    return header['files'][n]['content'] 
                raise Exception("No content")

        return _rec_get_item(self.header, item)

    def __contains__(self, item):
        def _rec_get_item(header, item):
            (p,n) = forward_path_split(item)
            if not 'files' in header:
                return False 
            if p != '':
                if not p in header['files'] or not 'files' in header['files'][p]:
                    return False;
                return _rec_get_item(header['files'][p], n)
            return True
        return _rec_get_item(self.header, item);

    def _extract_file(self, source, info, destination):
        """Locates and writes to disk a given file in the asar archive.

        Parameters
        ----------
        source : str
            Path of the file to write to disk
        info : dict
            Contains offset and size if applicable.
            If offset is not given, the file is assumed to be
            sitting outside of the asar, unpacked.
        destination : str
            Destination folder to write file into

        """

        r = info['content']
        dest = os.path.join(destination, source)
        with open(dest, 'wb') as f:
            f.write(r)

    def _extract_link(self, source, link, destination):
        """Creates a symbolic link to a file we extracted (or will extract).

        Parameters
        ----------
        source : str
            Path of the symlink to create
        link : str
            Path of the file the symlink should point to
        destination : str
            Destination folder to create the symlink into
        """
        dest_filename = os.path.normpath(os.path.join(destination, source))
        link_src_path = os.path.dirname(os.path.join(destination, link))
        link_to = os.path.join(link_src_path, os.path.basename(link))

        try:
            os.symlink(link_to, dest_filename)
        except OSError as e:
            if e.errno == errno.EXIST:
                os.unlink(dest_filename)
                os.symlink(link_to, dest_filename)
            else:
                raise e

    def _extract_directory(self, source, files, destination):
        """Extracts all the files in a given directory.

        If a sub-directory is found, this calls itself as necessary.

        Parameters
        ----------
        source : str
            Path of the directory
        files : dict
            Maps a file/folder name to another dictionary,
            containing either file information,
            or more files.
        destination : str
            Where the files in this folder should go to
        """
        dest = os.path.normpath(os.path.join(destination, source))

        if not os.path.exists(dest):
            os.makedirs(dest)

        for name, info in files.items():
            item_path = os.path.join(source, name)

            if 'files' in info:
                self._extract_directory(item_path, info['files'], destination)
            elif 'link' in info:
                self._extract_link(item_path, info['link'], destination)
            else:
                self._extract_file(item_path, info, destination)

    def extract(self, path):
        """Extracts this asar file to ``path``.

        Parameters
        ----------
        path : str
            Destination of extracted asar file.
        """
        if os.path.exists(path):
            raise FileExistsError()

        self._extract_directory('.', self.header['files'], path)

    @staticmethod
    def _reflat_header(header, target, buf, unpacked_files, unpacked_path):
        target['files'] = {}
        for (k,v) in header['files'].items():
            if "files" in v:
                target['files'][k] = {'files' : {}}
                Asar._reflat_header(v, target['files'][k], buf, unpacked_files, os.path.join(unpacked_path, k))
            elif ("content" in v):
                target['files'][k] = {}
                target['files'][k].update(v)

                if ('unpacked' in v):
                    fn = os.path.join(unpacked_path, k)
                    unpacked_files[fn] = v['content']
                    target['files'][k]['size'] = len(v['content'])
                else:
                    target['files'][k]['size'] = len(v['content'])
                    target['files'][k]['offset'] = str(buf.tell())
                    buf.write(v['content'])
                del target['files'][k]['content']


    def save(self, fn):
        target = {}
        buf = io.BytesIO()
        unpacked_files = {}
        self._reflat_header(self.header, target, buf, unpacked_files, "")

        data = Asar._build(target, buf.getvalue())
        data = data['fp'].getvalue()

        with open(fn, 'wb') as f:
            f.write(data)

        base_path = fn + ".unpacked"
        for (k,v) in unpacked_files.items():
            target = os.path.join(base_path, k)
            (p,n) = os.path.split(target)
            if not os.path.exists(p):
                os.makedirs(p)
            with open(target,"wb") as f:
                f.write(v)

    def mark_packed(self, item, state):
        def _rec_set_packed_item(header, item, state):
            (p,n) = forward_path_split(item)
            if not 'files' in header:
                raise Exception("error in header")
            if p != '':
                if not p in header['files'] or not 'files' in header['files'][p]:
                    raise Exception("item not found")
                return _rec_set_packed_item(header['files'][p], n, state)

            if 'content' in header['files'][n]: # First check if we already set this file
                if (state):
                    header['files'][n]['unpacked'] = True
                else:
                    header['files'][n].pop('unpacked',None)
            else:
                raise Exception("No content")
        return _rec_set_packed_item(self.header, item, state);

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.fp.close()




